from django.contrib.auth.models import Permission
from django.utils import timezone
from juntagrico.tests import JuntagricoTestCase

from juntagrico_assignment_request.models import AssignmentRequest


class AssignmentRequestTestCase(JuntagricoTestCase):
    @classmethod
    def setUpTestData(cls):
        # load from fixtures
        cls.load_members()
        cls.load_areas()
        cls.set_up_approvers()

    @classmethod
    def set_up_approvers(cls):
        cls.approver = cls.create_member('approver@email.org')
        cls.approver2 = cls.create_member('approver2@email.org')
        cls.approver.user.user_permissions.add(
            Permission.objects.get(codename='can_confirm_assignments'))
        cls.approver.user.save()
        cls.approver2.user.user_permissions.add(
            Permission.objects.get(codename='can_confirm_assignments'))
        cls.approver2.user.user_permissions.add(
            Permission.objects.get(codename='notified_on_unapproved_assignments'))
        cls.approver2.user.save()

    @classmethod
    def assignment_request_data(cls, approver=None, for_form=False, approved=False, **kwargs):
        if for_form:
            if approver is None:
                approver = ''
            else:
                approver = approver.pk
        date = timezone.now()
        data = {
            'job_time': date.strftime('%Y-%m-%d %H:%M') if for_form else date,
            'duration': 4,
            'amount': 0.5,
            'approver': approver,
            'activityarea': cls.area.pk if for_form else cls.area,
            'location': 'location',
            'description': 'description'
        }
        data.update(**kwargs)
        if not for_form:
            data['member'] = cls.member
            if approved:
                data['status'] = AssignmentRequest.CONFIRMED
        return data

    @classmethod
    def assignment_response_data(cls, decision='submit'):
        """
        :param decision: ['confirm', 'reject', 'submit'(default)]
        """
        return {
            'response': 'response',
            decision: True,
            'amount': 2,
            'duration': 5,
            'activityarea': cls.area.pk,
            'location': 'location'
        }



from django.test import override_settings

from .test_assignment_request import AssignmentRequestTests


@override_settings(ASSIGNMENT_UNIT='HOURS')
class AssignmentRequestHoursTests(AssignmentRequestTests):
    pass



# Generated by Django 4.2.9 on 2024-01-09 02:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('juntagrico_assignment_request', '0005_alter_assignmentrequest_amount'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='assignmentrequest',
            options={'permissions': (('can_confirm_assignments', 'Kann selbständige Arbeitseinsätze bestätigen'), ('can_confirm_assignments_for_area', 'Kann selbständige Arbeitseinsätze im eigenen Tätigkeitsbereich bestätigen'), ('notified_on_unapproved_assignments', 'Wird über nicht abgesprochene Arbeitseinsätze informiert')), 'verbose_name': 'Arbeitseinsatz Anfrage', 'verbose_name_plural': 'Arbeitseinsatz Anfragen'},
        ),
    ]



from juntagrico_assignment_request.models import AssignmentRequest
from . import AssignmentRequestTestCase


class ManipulationTests(AssignmentRequestTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.ar = AssignmentRequest.objects.create(**cls.assignment_request_data(cls.approver, approved=True))

    def test_job_type_manipulation(self):
        self.ar.assignment.job.type.activityarea = self.area2
        self.ar.save()
        self.assertEqual(self.ar.assignment.job.type.activityarea, self.area)

    def test_job_manipulation(self):
        self.ar.assignment.job.multiplier = 2
        self.ar.save()
        self.assertEqual(self.ar.assignment.job.multiplier, 1)

    def test_assignment_manipulation(self):
        self.ar.assignment.amount = 2
        self.ar.save()
        self.assertEqual(self.ar.assignment.amount, 0.5)



from juntagrico_assignment_request.models import AssignmentRequest

"""
This hack ensures that the jobs from assignment request can not be manipulated.
In the future juntagrico should offer ways to create custom job models in addons and make this hack obsolete.
"""


def bulk_save(elements):
    for element in elements:
        element.save()
        print(f'Resaved {element}')


def post_save_job_type(sender, instance, **kwds):
    """
    As we can not prevent admins from messing with autogenerated job types, we
    re-save all assignment requests. if any relevant part of this job type was changed,
    the requests will create a new matching job type and adapt their jobs to that.
    """
    # get assignment requests that are attached to any assignments of jobs of that job type and re-save them
    bulk_save(AssignmentRequest.objects.filter(
        assignment__job__in=instance.recuringjob_set.filter(assignment__assignmentrequest__isnull=False)
    ))


def post_save_recuringjob(sender, instance, **kwds):
    """
    same idea as post_save_job_type
    """
    # get assignment requests that are attached to any assignments of this job and re-save them
    bulk_save(AssignmentRequest.objects.filter(
        assignment__in=instance.assignment_set.filter(assignmentrequest__isnull=False)
    ))


def pre_save_assignment(sender, instance, **kwds):
    """
    If assignment comes from request, change back values as before saving it
    """
    if hasattr(instance, 'assignmentrequest'):
        ar = instance.assignmentrequest
        instance.member = ar.member
        instance.amount = ar.get_amount()
        instance.job = ar.get_matching_job()
        print(f'restored {instance} using {ar}')



from django.contrib.auth.models import Permission
from django.db.models import Q
from juntagrico.entity.member import Member


def get_approvers(general_only=False):
    codenames = ['can_confirm_assignments']
    if not general_only:
        codenames.append('can_confirm_assignments_for_area')
    perms = Permission.objects.filter(codename__in=codenames)
    return Member.objects.filter(
        Q(user__groups__permissions__in=perms) | Q(user__user_permissions__in=perms)
    ).distinct()


def all_notified_on_unapproved_assignments():
    perm = Permission.objects.get(codename='notified_on_unapproved_assignments')
    return Member.objects.filter(Q(user__groups__permissions=perm) | Q(user__user_permissions=perm)).distinct()



"""juntagrico_crowdfunding URL Configuration
"""
from django.urls import path
from juntagrico_assignment_request import views

app_name = 'juntagrico-assignment-request'
urlpatterns = [
    path('ar/assignment/request', views.request_assignment, name='request'),
    path('ar/assignment/delete/<int:request_id>/', views.delete_request_assignment, name='delete'),
    path('ar/assignment/edit/<int:request_id>/', views.edit_request_assignment, name='edit'),
    path('ar/assignment/list', views.list_assignment_requests, name='list'),
    path('ar/assignment/archive', views.list_archive, name='archive'),
    path('ar/assignment/respond/<int:request_id>/', views.respond_assignment_request, name='respond'),
    path('ar/assignment/confirm/<int:request_id>/', views.confirm_assignment_request, name='confirm'),
]



from django.conf import settings
from django.core.exceptions import ValidationError
from django.forms import ModelForm, DateTimeInput
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from django.urls import reverse

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit, HTML
from crispy_forms.bootstrap import FormActions

from juntagrico_assignment_request.models import AssignmentRequest
from juntagrico_assignment_request.utils import get_approvers


class DateTimeWidget(DateTimeInput):
    """ Widget using browsers date picker
    """
    input_type = 'datetime-local'

    def format_value(self, value):
        if isinstance(value, str):
            return value
        return value.strftime('%Y-%m-%dT%H:%M')


class AssignmentRequestForm(ModelForm):
    class Meta:
        model = AssignmentRequest
        fields = ('job_time', 'amount', 'duration', 'activityarea',
                  'approver', 'location', 'description')
        labels = {
            "amount": _("Anzahl Einsätze"),
            "approver": _("Abgesprochen mit"),
            "activityarea": _("Tätigkeitsbereich"),
        }

    def activityarea_queryset(self):
        qs = self.fields['activityarea'].queryset
        qs_filter = getattr(settings, 'ASSIGNMENT_REQUEST_AREAS', None)
        if callable(qs_filter):
            return qs_filter(qs)
        return qs

    def approvers_by_area(self):
        general_approvers = get_approvers(general_only=True)
        approvers = {
            area.id: area.coordinator.id
            for area in self.fields['activityarea'].queryset
            if area.coordinator in self.fields['approver'].queryset and area.coordinator not in general_approvers
        }
        return approvers

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['approver'].queryset = get_approvers()
        self.fields['activityarea'].queryset = self.activityarea_queryset()
        self.fields['duration'].widget.attrs['step'] = 1.0
        self.fields['job_time'].widget = DateTimeWidget()
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-md-3'
        self.helper.field_class = 'col-md-9'
        self.helper.layout = Layout(
            'job_time', 'amount', 'duration', 'activityarea',
            'approver', 'location', 'description',
            FormActions(
                Submit('submit', _('Absenden'), css_class='btn-success'),
            )
        )

    def clean(self):
        # check that approver is valid for the selected area
        cleaned_data = super().clean()
        valid_approvers = (get_approvers(general_only=True) |
                           get_approvers().filter(activityarea=cleaned_data['activityarea']))
        if cleaned_data['approver'] and cleaned_data['approver'] not in valid_approvers:
            raise ValidationError(_('Ungültige Auswahl bei "Abgesprochen mit"'))


class AssignmentResponseForm(ModelForm):
    class Meta:
        model = AssignmentRequest
        fields = ('amount', 'duration', 'activityarea', 'location', 'response')
        labels = {
            "amount": _("Anzahl Einsätze"),
            "activityarea": _("Tätigkeitsbereich"),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['duration'].widget.attrs['step'] = 1.0
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-md-3'
        self.helper.field_class = 'col-md-9'
        self.helper.layout = Layout(
            'amount', 'duration', 'activityarea', 'location', 'response',
            FormActions(
                Submit('confirm', _('Bestätigen'), css_class='btn-success'),
                Submit('reject', _('Ablehnen'), css_class='btn-danger'),
                Submit('submit', _('Nur Antwort senden'), css_class='btn-warning'),
                HTML('<a href="' + reverse('juntagrico-assignment-request:list') + '" class="btn">' + gettext("Abbrechen") + '</a>'),
            )
        )



from datetime import date

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.translation import gettext as _

from juntagrico.view_decorators import highlighted_menu

from juntagrico_assignment_request.forms import AssignmentRequestForm, AssignmentResponseForm
from juntagrico_assignment_request.mailer import membernotification, adminnotification
from juntagrico_assignment_request.models import AssignmentRequest


@login_required
@highlighted_menu('request_assignment')
def request_assignment(request, text_override=None):
    """
    Request an assignment
    """

    text = dict(
        request_sent=_("Deine Anfrage wurde erfolgreich verschickt.<br>"
                       "Du wirst per E-Mail benachrichtigt, sobald diese bestätigt wurde.")
    )
    text.update(text_override or {})

    member = request.user.member
    assignment_request_form = AssignmentRequestForm(request.POST or None)
    if request.method == 'POST' and assignment_request_form.is_valid():
        # Create request
        assignment_request = assignment_request_form.instance
        assignment_request.member = member
        assignment_request.save()
        adminnotification.request_created(assignment_request)
        messages.success(request, text['request_sent'])
        # do a redirect to avoid resending on reloading
        return redirect('juntagrico-assignment-request:request')

    renderdict = {
        'assignment_requests': AssignmentRequest.objects.filter(member=member).pending(),
        'form': assignment_request_form,
    }
    return render(request, "assignment_request/request.html", renderdict)


@login_required
def delete_request_assignment(request, request_id):
    """
    Delete an assignment request
    """

    assignment_request = get_object_or_404(AssignmentRequest, id=request_id)
    if assignment_request.member == request.user.member\
            and not assignment_request.assignment:
        assignment_request.delete()
    return redirect('juntagrico-assignment-request:request')


@login_required
@highlighted_menu('request_assignment')
def edit_request_assignment(request, request_id, text_override=None):
    """
    Edit and assignment request
    """

    text = dict(
        changed=_("Deine Änderungen wurden gespeichert."),
        rerequested=_("Deine Änderungen wurden gespeichert und zur (erneuten) Bestätigung verschickt.")
    )
    text.update(text_override or {})

    member = request.user.member
    assignment_request = get_object_or_404(AssignmentRequest, id=request_id, member=member)
    old_amount = assignment_request.get_amount()
    old_approver = assignment_request.approver
    assignment_request_form = AssignmentRequestForm(request.POST or None, instance=assignment_request)
    if request.method == 'POST' and assignment_request_form.is_valid():
        # edit request
        new_request = assignment_request_form.instance
        if new_request.get_amount() > old_amount:
            # if significant changes were made re-approval is needed
            new_request.status = AssignmentRequest.REQUESTED
            new_request.save()
            adminnotification.request_changed(new_request)
            messages.success(request, text['rerequested'])
        else:
            if new_request.status is not AssignmentRequest.REQUESTED:
                # keep approver, if they replied already
                new_request.approver = old_approver
            new_request.save()
            messages.success(request, text['changed'])
        return redirect('juntagrico-assignment-request:request')

    renderdict = {
        'form': assignment_request_form,
    }
    return render(request, "assignment_request/edit.html", renderdict)


@permission_required('juntagrico_assignment_request.can_confirm_assignments')
def list_assignment_requests(request):
    """
    List assignment requests
    """
    ar = AssignmentRequest.objects.filter(status=AssignmentRequest.REQUESTED).for_approver(request.user.member)
    renderdict = {
        'assignment_requests': ar,
    }
    return render(request, "assignment_request/list.html", renderdict)


@permission_required('juntagrico_assignment_request.can_confirm_assignments')
def list_archive(request):
    ar = AssignmentRequest.objects.exclude(status=AssignmentRequest.REQUESTED).for_approver(request.user.member)
    renderdict = {
        'assignment_requests': ar,
        'archive': True
    }
    return render(request, "assignment_request/list.html", renderdict)


@permission_required('juntagrico_assignment_request.can_confirm_assignments')
def respond_assignment_request(request, request_id, text_override=None):
    """
    Confirm or reject an assignment request
    """

    text = dict(
        replied=_("Antwort gesendet."),
        already_replied=_("Die Anfrage wurde bereits beantwortet.")
    )
    text.update(text_override or {})

    assignment_request = get_object_or_404(AssignmentRequest, id=request_id)
    if assignment_request.status != AssignmentRequest.REQUESTED:
        messages.warning(request, text['already_replied'])
        return redirect('juntagrico-assignment-request:list')

    assignment_response_form = AssignmentResponseForm(request.POST or None, instance=assignment_request)
    if request.method == 'POST' and assignment_response_form.is_valid():
        assignment_request = assignment_response_form.instance
        assignment_request.response_date = date.today()
        if 'confirm' in request.POST:
            assignment_request.status = AssignmentRequest.CONFIRMED
        elif 'reject' in request.POST:
            assignment_request.status = AssignmentRequest.REJECTED
        adminnotification.request_handled_by_other_approver(assignment_request, request.user.member)
        assignment_request.approver = request.user.member
        assignment_request.save()
        membernotification.request_handled(assignment_request)
        messages.success(request, text['replied'])
        return redirect('juntagrico-assignment-request:list')

    renderdict = {
        'assignment_request': assignment_request,
        'form': assignment_response_form,
    }
    return render(request, "assignment_request/respond.html", renderdict)


@permission_required('juntagrico_assignment_request.can_confirm_assignments')
def confirm_assignment_request(request, request_id, text_override=None):
    """
    Confirm an assignment request directly
    """
    text = dict(
        confirmed=_("Anfrage bestätigt."),
        already_confirmed=_("Anfrage bereits bestätigt.")
    )
    text.update(text_override or {})

    assignment_request = get_object_or_404(AssignmentRequest, id=request_id)
    if not assignment_request.is_confirmed():
        assignment_request.response_date = date.today()
        assignment_request.response = ''
        assignment_request.status = AssignmentRequest.CONFIRMED
        adminnotification.request_handled_by_other_approver(assignment_request, request.user.member)
        # overwrite approver in any case to actual approver
        assignment_request.approver = request.user.member
        assignment_request.save()
        membernotification.request_handled(assignment_request)
        messages.success(request, text['confirmed'])
    else:
        messages.warning(request, text['already_confirmed'])
    return redirect('juntagrico-assignment-request:list')



# Generated by Django 4.2.9 on 2024-01-08 20:06

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('juntagrico_assignment_request', '0004_auto_20211207_1101'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assignmentrequest',
            name='amount',
            field=models.FloatField(default=1.0, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Wert'),
        ),
    ]



from django.apps import AppConfig
from django.db.models import signals


class JuntagricoAssignmentRequestAppconfig(AppConfig):
    name = "juntagrico_assignment_request"
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):
        from juntagrico.util import addons
        addons.config.register_version(self.name)

        # Connect signals
        from juntagrico_assignment_request import models, hack
        from juntagrico.entity.jobs import JobType, RecuringJob, Assignment
        signals.pre_save.connect(models.AssignmentRequest.pre_save, sender=models.AssignmentRequest)
        signals.post_save.connect(hack.post_save_job_type, sender=JobType)
        signals.post_save.connect(hack.post_save_recuringjob, sender=RecuringJob)
        signals.pre_save.connect(hack.pre_save_assignment, sender=Assignment)

        #############################################
        # Hide automatically generated jobs and types
        # in admin by monkey patching
        from juntagrico.admins.job_admin import JobAdmin
        from juntagrico.admins.job_type_admin import JobTypeAdmin

        def get_job_type_queryset(self, request):
            return original_get_job_type_queryset(self, request).exclude(
                recuringjob__assignment__assignmentrequest__isnull=False)

        def get_job_queryset(self, request):
            return original_get_job_queryset(self, request).exclude(
                assignment__assignmentrequest__isnull=False)

        original_get_job_type_queryset = JobTypeAdmin.get_queryset
        JobTypeAdmin.get_queryset = get_job_type_queryset
        original_get_job_queryset = JobAdmin.get_queryset
        JobAdmin.get_queryset = get_job_queryset
